/***************************************************************************//**
 * @file    night-light-PIR.ino
 * @brief   Arduino program for automatic night light.
 *
 * @description
 * Night light (hallway light), with PIR, daylight sensor, programmable LEDs
 * and many run-time settings.
 * Arduino program.
 *
 * Connections (Arduino pins):
 * -  D2   Button down (version 1 only)
 * -  D3   Button up (version 1 only)
 * -  D4   Rotary encoder SW
 * -  D5   Rotary encoder A  (A, B swapped on cheap encoder?)
 * -  D6   Rotary encoder B
 * -  D8   LED string (ring) with WS2812 (data-in of string)
 * -  D13  LED reflecting PIR output
 * -  D13  LED that flashes during setting mode
 * -  A0   PIR output
 * -  A3   Daylight sensor (analog)
 *
 * Platform: Arduino Pro Mini 16MHz 5V ATmega168.
 *
 * Arduino libraries used:
 * - EEPROM (built-in)
 * - SoftTimer (Volker)  @n
 *   http://volker.top.geek.nz/arduino/
 * - PushButton (Volker)  @n
 *   http://volker.top.geek.nz/arduino/
 * - Adafruit_NeoPixel (Adafruit)  @n
 *   https://github.com/adafruit/Adafruit_NeoPixel
 * - Streaming (Mikal Hart) (for debugging only)  @n
 *   http://arduiniana.org/libraries/streaming/
 *
 * See main page for user manual and further information (end of this file).
 *
 * @TODO
 * - Empty.
 *
 * @author Volker Kuhlmann  @n
 * http://volker.top.geek.nz/contact.html
 *
 * @download
 * http://volker.top.geek.nz/arduino/
 *
 * @copyright
 * Copyright (C) 2016 by Volker Kuhlmann  @n
 *               http://volker.top.geek.nz/contact.html  @n
 * All rights reserved.
 * Released under the terms of the GNU General Public License (GPL) Version 3
 * or (at your option) any later version.  @n
 * See https://www.gnu.org/licenses/ for details.
 *
 * @history
 * - v1.0, 19+20 Aug 2016  Volker Kuhlmann
 *   - Initial version. For development board.
 *   - PIR, up/down button for setting on-brightness, ramping light up+down.
 *   - Storing brightness in EEPROM TBI.
 *   - Program size: appr 4k, 5.5k with debugging.
 * - v1.1, 21 Aug 2016  Volker Kuhlmann
 *   - Store brightness setting in EEPROM.
 *   - Flash Arduino LED while in setting mode.
 * - v1.2, 23 Aug 2016  Volker Kuhlmann
 *   - Store on-duration in EEPROM; changeable by setting.
 *   - Fix button debouncing.
 *   - Show a colour pattern on start-up.
 * - v1.3, 23 Aug 2016  Volker Kuhlmann
 *   - Rename from hall-light-PIR to night-light-PIR.
 *   - Add daylight sensor.
 *   - Change pin assignment from devel board to final assembly.
 *   - TODO: Fine-tune the compile-time defaults.
 * - v1.4, 07 Sep 2016  Volker Kuhlmann
 *   - Specify LED to use for index=0 (useful for LED rings).
 *   - Use SoftTimer and PushButton libraries.
 *   - Add application summary text, printed to serial on start-up.
 *   - Add rotary encoder, and use it to set on-brightness and on-duration.
 * - v1.5, 11 Sep 2016  Volker Kuhlmann
 *   - Change duration of reaching full brightness and off at runtime.
 *   - Change heartbeat brightness and light colour at runtime.
 *   - Code clean-ups.
 * - v1.6, 19 Sep 2016  Volker Kuhlmann
 *   - Doxygen documentation.
 *   - Remove INT_LIT_ARRAY_ALT alternative.
 *   - Re-order source code parts. Rename some classes.
 *   - Ignore PIR activation immediately after power-on (false triggers).
 *   - Set the ambient light level threshold for day/night at runtime.
 *   - Improve noise filtering for ambient light measurements, and reduce
 *     indicator light and setting LED brightness.
 *******************************************************************************
 */

//! Program version. The number is extracted automatically for the doxyfile.
#define VERSIONINFO "v1.6, 19 Aug 2016 - 19 Sep 2016"


/*
 * Debugging / Development, Macros for Configuration
 */

// When commenting out a macro, comment out the comment as well,
// or doxygen gets confused about what the comment belongs to.

//! Write debugging output to serial port if defined (costs ROM + RAM!).
#define DEBUG_APP
//
////! Write ADC value of light sensor to serial port 3/s if defined.
////! Only with DEBUG_APP.
//#define DEBUG_LIGHT_ADC

//! @brief Show application version on serial port if defined.
//
//! (This opens serial port and costs ROM + RAM!)
#define SHOW_APP_VERSION

//! @brief Show file name with application version.
//
//! (Contains full path when using Arduino IDE!)
#define SHOW_APP_VERSION_FILE

////! Show start-up blinks if defined (to indicate a reset has occurred).
//#define SHOW_START_BLINKS

//! Support up/down push buttons (in addition to rotary encoder).
#define SUPPORT_UPDOWN_BUTTONS


/*
 * Headers, Arduino Libraries
 */

// Doxygen does not comment #include's and gets confused with comments here.

// AVR EEPROM (built-in)
#include <EEPROM.h>

// Streaming (Mikal Hart) (Make outputing to serial port usable).
#ifdef DEBUG_APP
#include <Streaming.h>
#endif

// Software timers (Volker)
#include <SoftTimer.h>

// Switches, rotary encoder (Volker)
#include <PushButton.h>
#ifdef DEBUG_APP
/* Print new state of a push button.
 * The library calls this, but the user must implement it.
 * (Doxygen won't include this function in the documentation.)
 */
void PushButton_show::print(char status)
{
     Serial << "Button " << _id << ' ' << status << endl;
}
#endif

// WS2812 LED strings (Adafruit)
#include <Adafruit_NeoPixel.h>


/*
 * Pins, Hardware Configuration
 */

/*! @brief Pin assignments and hardware configuration.
 *
 * Digital pins 1-13, LED_BUILTIN is normally 13.
 * Analog pins A0-A7.
 *
 * Grouped in a class to keep it together in the documentation.
 */
class HW
{
  public:
    // Rotary encoder
    static const byte PIN_ROTENC_A = 5;   //!< Rotary encoder A output
    static const byte PIN_ROTENC_B = 6;   //!< Rotary encoder B output
    static const byte PIN_ROTENC_SW = 4;  //!< Rotary encoder push switch
    static const byte ROTENC_COUNTS_PER_CLICK = 4;
                                          //!< Rotary encoder counts per click.

#ifdef SUPPORT_UPDOWN_BUTTONS
    //! @brief Down button
    //! Further models will only have a rotary encoder.
    static const byte PIN_BUTTON_D = 2;
    //! @brief Up button
    //! Further models will only have a rotary encoder.
    static const byte PIN_BUTTON_U = 3;
#endif

    //! LED string (ring) WS2812 data in
    static const byte PIN_LEDSTRIP = 8;
    //! LED string (ring) length.
    //
    //! Length must be at least as big as Light::on_time_choice_size!
    static const byte LEDSTRIP_LENGTH = 12;
    //! Type of LED (colour order, controller speed).
    static const int LEDSTRIP_TYPE = NEO_GRB + NEO_KHZ800;

    //! PIR sensor (motion detector). Digital signal.
    static const byte PIN_PIR = A0;

    //! Daylight sensor. Analog input!
    static const byte PIN_DAYLIGHT_SENSOR = A3;

    //! LED reflecting PIR output
    //
    //! If this is shared with PIN_STATUS_SETMODE, PIN_STATUS_SETMODE takes
    //! precedence while in setting mode.
    //! (Built-in LED is normally pin 13.)
    static const byte PIN_STATUS_PIR = LED_BUILTIN;

    //! LED showing setting mode.
    //
    //! May be shared with PIN_STATUS_PIR.
    static const byte PIN_STATUS_SETMODE = LED_BUILTIN;

    //! Baud rate for serial port.
    static const unsigned long RS232_speed = 115200;
    //! Mode for serial port.
    static const byte RS232_mode = SERIAL_8N1;
};


/*
 * Settings (saved in non-volatile memory)
 */

/*!
 * @brief Settings that can be changed at run-time.
 *
 * Stored in non-volatile memory.
 */
class SettingValues
{
  public:
    /* In decreasing order of importance: */

    //! Light brightness (0..255).
    byte brightness;

    //! Light on time after trigger (0..Light::on_time_choice_size).
    byte on_time_index;

    //! On ramp length (0..Light::ramp_time_choice_size).
    byte on_ramp_index;

    //! Off ramp length (0..Light::ramp_time_choice_size).
    byte off_ramp_index;

    //! Heartbeat brightness (0..Light::heartbeat_brightness_max).
    byte hb_brightness;

    //! Colour wheel position (0..255).
    byte wheel_position;

    //! Colour wheel level (0..255).
    //
    //! Level 0: RGB colour wheel.
    //! Level 255: CMY colour wheel.
    //! Level 128: White.
    //! For off-white colours try levels around 90 or 170.
    byte wheel_level;

    //! Light colour (24bit RGB value).
    uint32_t colour;

    //! ADC value below which "night" is assumed. 10 bit.
    uint16_t night_threshold_ADC;

    /* Other ideas: */
    // Random lighting at night, in "away" mode. The light is unlikely to be
    // bright enough for this.

    //! Index into setup menu.
    //
    //! Not a setting value, but stored here for use by setting up functions.
    //! Side-effect: Menu is entered at location where it was last exited from.
    byte menu_index;

    //! Settings checksum.
    byte checksum;
};

/*! @brief Default values for all the settings.
 *
 * In the order as in SettingValues!
 * @note The compiler does not check for any errors here!
 */
static const SettingValues SettingsDefaults {
  50,       // On-brightness.
  1,        // On-time, index in array of seconds.
  1,        // On-ramp time, index in array of seconds.
  3,        // Off-ramp time, index in array of seconds.
  3,        // Heartbeat brightness.
  0,        // Colour wheel position.
  128,      // Colour wheel level.
  0xFFB070UL, // Light colour (any value - ignored, calculated from pos+level)
  150,      // Night brightness threshold. ADC value.

  0,        // Menu index (any value - ignored).
  0         // Checksum location (any value - ignored).
};

//! Settings that can be changed at run-time.
//
//! Must fit into uint8_t!
enum setting_item {
  SET_NONE, 	    	//!< Not in setting mode.
  SET_MENU, 	    	//!< In setting menu (colour idxcolour_menu).

  // These are the settings that can be changed at run-time.
  // They're indexed by the settings menu in order, starting with the first.
  SET_BRIGHTNESS,   	//!< Set brightness.
  SET_ON_TIME,	    	//!< Set time the light is on for. Index.
  SET_ON_RAMP_TIME, 	//!< Set duration to reach full brightness.
  SET_OFF_RAMP_TIME,	//!< Set duration to reach fully off.
  SET_HB_BRIGHTNESS,	//!< Set heartbeat brightness.
  SET_COLOUR_POS,   	//!< Set colour wheel position.
  SET_COLOUR_LVL,   	//!< Set colour wheel level.
  SET_AMBIENT_BRIGHTNESS,
                      //!< Set ambient brightness, don't turn light on above it.

  SET_MENU_FIRST = SET_BRIGHTNESS,  //!< Same as first setting item.
  SET_NUM_ITEMS = 8 	//!< @brief Number of settings, for settings menu.
        //!<  Ensure this matches the number of items above!
};


/*
 * Data and object types
 */

/*!
 * @brief Status display object.
 *
 * Show status information.
 * Currently:
 * - The status of the PIR output.
 * - Whether the system is in setting mode.
 */
class Statusdisplay
{
  public:
    //! @brief Speed of setting-mode LED flashing while in setting mode.
    //! Half of a period, in milliseconds.
    static const uint16_t SETTINGFLASH_MS = 100;
    //! Speed of flashing while in setting mode for SET_AMBIENT_BRIGHTNESS.
    //
    //! The period is 2 * SETTINGFLASH_MS.
    static const uint8_t SETTINGFLASH_SHORT_MS = 3;

    //! Set up the status display.
    void setup();
    //! Poll the status display.
    void loop();

    //! Show whether the PIR sensor detected motion.
    void PIRactive(bool isactive);
    //! Show whether setting mode was entered. Call on entry and exit.
    void setting_mode(uint8_t mode);

  protected:

  private:
    //! Show an active setting mode.
    uint8_t _active_settingmode;
    //! Timer for setting mode LED flashing.
    SoftTimer _timer_settingmode;
    //! State of the LED while in setting mode.
    bool _led_setmode;
};

/*!
 * @brief PIR sensor (motion detector) object.
 *
 * Handle the PIR sensor (motion detector).
 */
class PIRsensor
{
  public:
    //! Ignore detector activation within this long after power-on.
    //! Milliseconds.
    static const uint16_t START_IGNORE_MS = 3000;

    //! Set up the PIR sensor.
    void setup();
    //! Poll the PIR sensor.
    void loop();

    //! Whether the PIR sensor detected any motion.
    bool is_active();
    //! @brief Whether motion was detected. Unprocessed sensor output.
    //! Actual pin state (active high).
    bool raw();
};

/*!
 * @brief Daylight sensor object.
 *
 * Handle the daylight sensor.
 */
class Daylightsensor
{
  public:
    //! Set up the daylight sensor.
    void setup();
    //! Poll the daylight sensor.
    void loop();

    //! Current level of the light sensor. 10 bit value.
    //
    //! Noise filtering is performed.
    word light_value();
    //! Whether it is currently night.
    bool is_night();

    //! Last read ADC value. No filtering, For debugging.
    uint16_t ADCval();

  protected:
    //! ADC sample interval.
    static const byte _sampleperiod_ms = 22;

    //! Last filtered ADC value of the light sensor.
    SoftTimer _timer_adc;

  private:
    //! Size of buffer for noise filtering.
    static const byte _bufsize = 32;
    //! Buffer for ADC values. Used as ring buffer.
    word _buf[_bufsize];
    //! Last/current index into buffer.
    byte _idx;
};

/*!
 * @brief Light object.
 *
 * Handle the programmable LED string (ring).
 * The LEDs are used as light, and as display in setting mode.
 */
class Lightring
{
  public:
    //! Length of time to show the start-up indicator for.
    //
    //! 0 turns it off (and saves a bit of code too).
    static const unsigned int start_indicator_ms = 1000;

    //! Minimum value the brightness can be set to.
    //
    //! Can't be turned completely off to avoid confusion.
    static const byte on_brightness_min = 10;
    //! Maximum value the brightness can be set to.
    static const byte on_brightness_max = 200;
    //! Brightness change with each step.
    static const byte on_brightness_change = 10;

    //! Minimum value the heartbeat brightness can be set to.
    //
    //! 0 turns it off.
    static const byte heartbeat_brightness_min = 0;
    //! Maximum value the heartbeat brightness can be set to.
    static const byte heartbeat_brightness_max = 30;
    //! Brightness change with each step.
    static const byte heartbeat_brightness_change = 1;
    //! Heartbeat light colour.
    static const unsigned long int heartbeat_colour = 0xFF0000;
    //! Time each heartbeat LED is on, before moving to the next. Milliseconds.
    static const unsigned int heartbeat_interval_ms = 1000;
    //! Heartbeat won't show for the first this many milliseconds.
    static const unsigned int heartbeat_delay_ms = 3000;

    //! Brightness when showing an index with the programmable light.
    static const byte index_brightness = 20;
    //! Colour for the setup menu index.
    static const unsigned long int idxcolour_menu = 0x0000FF;
    //! Colour for the item value index.
    static const unsigned long int idxcolour_value = 0x00FF00;
    //! The LED in the ring which is used for the lowest index.
    //
    //! This allows to mount the ring at any angle.
    static const byte index_offset = 4; // which LED to use for lowest index

    //! Colour to use when setting ambient light day/night threshold.
    //
    //! At index brightness.
    static const uint32_t ambient_colour = 0x532900;

    //! Number of choices for the light's on-time.
    static const byte on_time_choice_size = 8;
    //! Choices for the length of time the light is on for. Seconds.
    // C++ doesn't allow static const arrays that can't be indexed at compile
    // time.
    // Works (not more code than const array outside of class).
    // Downside: function, not array; cumbersome.
    static const uint16_t get_on_time_choice(byte index) {
      static const uint16_t on_time_choice_lcl[] = {
        15, 30, 45, 60, 2*60, 4*60, 8*60, 16*60
      };
      return on_time_choice_lcl[index];
    }

    //! Number of choices for the speed of turning on/off.
    static const byte ramp_time_choice_size = 6;
    //! Duration (ms) over which the light is progressiveley turned on/off.
    static const uint16_t get_ramp_time_choice_ms(byte index) {
      static const uint16_t ramp_time_choice_ms_lcl[] = {
        1*1000, 2*1000, 4*1000, 7*1000, 10*1000, 15*1000
      };
      return ramp_time_choice_ms_lcl[index];
    }

    //! Set up the LED ring.
    void setup();
    //! Poll the LED ring.
    void loop();

    //! Set the light on or off. The light switch!
    void setlight(bool onoff);

    //! Use light to show that setup mode is entered, for the type given.
    void setup_begin(byte setting_type);
    //! Setup mode is finished.
    void setup_end();

    //! Colour wheel, morphing from RGB over white to CMY.
    static uint32_t ColourWheelRGBWCMY(uint8_t index, uint8_t level);

  protected:
    //! LED string (ring) object.
    //
    //! Initialising it in setup() saves 716 bytes code! (Why?)
    Adafruit_NeoPixel strip;

    //! Brightness the light is set to by _show_light().
    byte _brightness = 0;
    //! Brightness the light should reach (0 for off).
    byte _target_brightness = 0;
    //! The state of the light (on, off). "Light switch".
    bool _turn_on;
    //! Whether light brightness is currently ramped up or down.
    bool _ramping;

    //! Change the actual towards the target brightness, if they differ.
    void _ramp();
    //! Set the light to the current brightness and colour.
    void _show_light();
    //! Show an index; different colour for setup menu and setup items.
    void _show_index(byte index);
    //! Show a colour wheel, using position and level from the settings.
    void _show_wheel();
    //! Show the heart beat.
    void _show_heartbeat();
    //! Show indication that ambient light levels are measured.
    void _show_ambient();

  private:
    //! Timer used for ramping.
    SoftTimer _timer_ramp;

    //! LED which is currently on for the heart beat.
    byte _heartbeat_nr = 0;
    //! Timer used for the heart beat.
    SoftTimer _timer_heartbeat;

    //! Setup item to show when in setting mode.
    byte _in_setup = SET_NONE;

    //! When displaying an index, the last number displayed.
    byte _last_index = -1;
    //! Flashing the index display; slower for higher index values.
    bool _idx_pause;
    //! Timer used for flashing the index display.
    SoftTimer _timer_idx_flash;
};

/*!
 * @brief Non-volatile (NV) storage for settings.
 *
 * Load and store settings to NV memory, and calculate/check checksums.
 * setup() and loop() don't make sense here.
 */
class NVSettings
{
  public:
    //! Address at which the settings are stored in EEPROM.
    static const unsigned int EEPROMaddr = 2;
    //! Number of bytes of settings data, including checksum.
    static const byte valsize = sizeof(SettingValues);

    //! All the settings.
    SettingValues val;

    //! Load the settings from NV memory.
    void load();
    //! Save the settings to NV memory.
    void store();
    //! Calculate checksum of current data in val and check if correct.
    bool is_good_checksum();
    //! Update the checksum value in val to give a correct checksum for val.
    void checksum_update();

  protected:
    //! The correct checksum for the data.
    static const byte _good_checksum_value = 0xC5;

    //! Most recently calculated checksum value.
    byte _checksum;

    //! Dump EEPROM settings storage. For debugging.
    void _EEPROMdump(char id); // debugging
};

/*!
 * @brief Change settings at run-time.
 *
 * Act on button input and change settings accordingly.
 * Use the light as display.
 */
class Settingup
{
  public:
    //! Exit setting mode after this much inactivity. In milliseconds.
    static const uint16_t setup_timeout_ms = 4000;

    //! Set up the setup object.
    void setup();
    //! Poll the setup object.
    void loop();

    //! Whether system is in setting mode, or normal operation.
    bool in_setup_mode();

  protected:
    //! Whether system is in setup mode, and which setup item. State variable.
    byte _setting_up = SET_NONE;

    //! Check user input for whether to enter setting mode.
    void _check_enter_setup();
    //! Check whether to leave setting mode.
    void _check_exit_setup();
    //! Show the newly selected settings value. Debugging only.
    void _exit_debug();

    //! Set a 1-of-N choice. Minimum is 0.
    byte _set_range(byte current, byte max);
    //!  Set a 1-of-N choice.
    byte _set_range(byte current, byte min, byte max);
    //! Set the light colour. Actual selected colour is shown.
    void _set_colour();
    //! Set current ambient light level as threshold for day/night.
    void _set_ambient_night();
    //! Set the light's brightness, showing actual selected value.
    void _set_brightness();
    //! Set for how long the light is turned on after trigger.
    void _set_on_time();

  private:
    //! Timer used to exit setting mode after inactivity.
    SoftTimerExt _setup_timeout;

#ifdef SUPPORT_UPDOWN_BUTTONS
    //! Flag whether setting mode was entered by up/down buttons.
    bool _entered_by_updown;
#endif
};


/*
 * Global variables
 */

#define STRINGIFY(s) # s //!< Turn the parameter into a string.
#define ARDUINO_EXPAND(arg) STRINGIFY(arg) //!< Expand arg value into string.
#define ARDUINO_STR ARDUINO_EXPAND(ARDUINO) //!< Arduino version, as string.
//
//! String showing that the compiler used LTO. Passed in from Makefile.
//
//! Arduino 1.6.10, gcc 4.9, uses it by default.
#ifdef LTO
#define LTO_STR " LTO"
#else
#define LTO_STR
#endif
//
//! Summary info (about-text).
PROGMEM const char APPLICATIONSUMMARY_TEXT[] =
  "\r\nNight-light " VERSIONINFO " (C) by Volker Kuhlmann\r\n"
  "All rights reserved.\r\n"
  "avr libc " __AVR_LIBC_VERSION_STRING__ ", gcc " __VERSION__ LTO_STR
  ", Arduino " ARDUINO_STR "\r\n"
#ifdef SHOW_APP_VERSION_FILE
  "File: " __FILE__ "\r\n\r\n"
#endif
;

//! Flash strings. Similar to F(), but for PROGMEM addresses, not literals.
//
//! This allows using strings in progmem with the print functions.
#define FS(str) (reinterpret_cast<const __FlashStringHelper *>(str))

//! Status display object, instantiated.
Statusdisplay statusdisp = Statusdisplay();

//! PIR sensor object, instantiated.
PIRsensor pir = PIRsensor();

//! Daylight sensor object, instantiated.
Daylightsensor daylight = Daylightsensor();

//! Light ring object, instantiated.
Lightring light = Lightring();

//! NV storage object, instantiated.
NVSettings settings = NVSettings();

//! Setup object, instantiated.
Settingup settingup = Settingup();

#ifdef DEBUG_APP
//! Rotary encoder with push switch object, instantiated. Verbose mode.
RotaryEncoder_show encoder = RotaryEncoder_show(HW::PIN_ROTENC_A,
  HW::PIN_ROTENC_B, HW::PIN_ROTENC_SW, HW::ROTENC_COUNTS_PER_CLICK, 'E');
#ifdef SUPPORT_UPDOWN_BUTTONS
//! Down-button object, instantiated. Verbose mode.
PushButton_show button_down = PushButton_show(HW::PIN_BUTTON_D, 'D');
//! Up-button object, instantiated. Verbose mode.
PushButton_show button_up = PushButton_show(HW::PIN_BUTTON_U, 'U');
#endif
#else
//! Rotary encoder with push switch object, instantiated.
RotaryEncoder encoder = RotaryEncoder(HW::PIN_ROTENC_A, HW::PIN_ROTENC_B,
  HW::PIN_ROTENC_SW, HW::ROTENC_COUNTS_PER_CLICK);
#ifdef SUPPORT_UPDOWN_BUTTONS
//! Down-button object, instantiated.
PushButton button_down = PushButton(HW::PIN_BUTTON_D);
//! Up-button object, instantiated.
PushButton button_up = PushButton(HW::PIN_BUTTON_U);
#endif
#endif


/*
 * Object class methods
 */

/*!
 * @return void
 */
void Statusdisplay::setup()
{
  pinMode(HW::PIN_STATUS_PIR, OUTPUT);
  pinMode(HW::PIN_STATUS_SETMODE, OUTPUT);
}

/*!
 * Flash the setting mode LED while in setting mode.
 * Shorter on-time for SET_AMBIENT_BRIGHTNESS to avoid interference with
 * ambient light levels.
 * @return void
 */
void Statusdisplay::loop()
{
  if (SETTINGFLASH_SHORT_MS != SETTINGFLASH_MS) {
    if (_active_settingmode != SET_NONE && _timer_settingmode.is_expired()) {
      _led_setmode = ! _led_setmode;
      digitalWrite(HW::PIN_STATUS_SETMODE, _led_setmode);

      // Period is 2*SETTINGFLASH_MS. On-time is SETTINGFLASH_MS, or
      // SETTINGFLASH_SHORT_MS when setting SET_AMBIENT_BRIGHTNESS.
      uint8_t waitms = 2 * SETTINGFLASH_MS - SETTINGFLASH_SHORT_MS;
      if (_led_setmode)
        waitms = SETTINGFLASH_SHORT_MS;
      if (_active_settingmode != SET_AMBIENT_BRIGHTNESS)
        waitms = SETTINGFLASH_MS;
      _timer_settingmode.startN(waitms);
    }
  }
  else if (_active_settingmode && _timer_settingmode.interval(SETTINGFLASH_MS)) {
        _led_setmode = ! _led_setmode;
        digitalWrite(HW::PIN_STATUS_SETMODE, _led_setmode);
  }
}

/*! @brief
 *
 * If the LEDs for setting mode and PIR status are one and the same,
 * show only setting mode.
 * This is meant to show the state of the PIR sensor output signal, without any
 * filtering.
 * @param isactive bool True if motion was detected.
 * @return void
 */
void Statusdisplay::PIRactive(bool isactive)
{
  // Setting mode display takes precedence.
  if (HW::PIN_STATUS_SETMODE == HW::PIN_STATUS_PIR
      && _active_settingmode)
    return;

  digitalWrite(HW::PIN_STATUS_PIR, isactive);
}

/*! @brief
 *
 * Also call with SET_NONE when exiting setting mode.
 * @param mode The setting mode that was entered.
 * @return void
 */
void Statusdisplay::setting_mode(uint8_t mode)
{
  _active_settingmode = mode;
  _led_setmode = false;
}

/*! @brief
 *
 * @return void
 */
void PIRsensor::setup()
{
  // Don't use pull-up for active-high signal.
  pinMode(HW::PIN_PIR, INPUT);
}

/*! @brief
 *
 * @return void
 */
void PIRsensor::loop()
{
}

/*! @brief
 *
 * The state of the PIR sensor output signal.
 * @return bool True if sensor detected motion.
 */
bool PIRsensor::raw()
{
  // PIR output is active high.
  bool out = digitalRead(HW::PIN_PIR);
  return out;
}

/*! @brief
 *
 * May contain further processing of the PIR signal.
 * Ignore an active state within a few seconds after power-on, while the PIR
 * controller stabilises.
 *
 * @return bool True if the sensor detected motion.
 */
bool PIRsensor::is_active()
{
  if (millis() < START_IGNORE_MS)
    return false;
  return raw();
}

/*! @brief
 *
 * @return void
 */
void Daylightsensor::setup()
{
  pinMode(HW::PIN_DAYLIGHT_SENSOR, INPUT);

  // ADC reference. DEFAULT = Vcc, INTERNAL = 1.1V (328P).
  // 1.1V is too low to be much use, Vcc will give noise.
  analogReference(DEFAULT);
}

/*! @brief
 *
 * @return void
 */
void Daylightsensor::loop()
{
  // Periodically read the ADC into a ring buffer.
  // This could be programmed better with fewer resources, e.g. stepping
  // the result by 1 in the direction of the reading with each period.
  // But we have the resources. Buffer size 32 with 22ms keeps numbers stable,
  // as long as there is no change in light, like people moving.
  if (_timer_adc.interval(_sampleperiod_ms)) {
    // Ignore that _idx may be uninitialised, but range-check before use.
    ++_idx;
    if (_idx >= _bufsize)
      _idx = 0;
    _buf[_idx] = (analogRead(HW::PIN_DAYLIGHT_SENSOR));

#if defined(DEBUG_LIGHT_ADC) && defined(DEBUG_APP)
    static Timer serout;
    if (serout.interval(333)) {
      Serial << _buf[_idx] << '\t' << light_value() << endl;
    }
#endif
  }
}

/*! @brief
 *
 * Values increase with increasing brightness. 10 bit range.
 * @return word Filtered ADC value of daylight sensor.
 */
word Daylightsensor::light_value()
{
  // Ignore that buffer contains random values after start-up; that issue
  // disappears after _bufsize samples (_bufsize * _sampleperiod_ms).
  unsigned long avg = 0;
  for (byte i = 0; i < _bufsize; ++i)
    avg += _buf[i];
  avg /= _bufsize;
  return avg;
}

/*! @brief
 *
 * @return bool True if it dark enough to turn the light on.
 */
bool Daylightsensor::is_night()
{
  return light_value() <= settings.val.night_threshold_ADC;
}

/*! @brief
 *
 * @return uint16_t Last read ADC value, no filtering. Development only.
 */
uint16_t Daylightsensor::ADCval()
{
  return _buf[_idx];
}

/*! @brief
 *
 * After initialising the hardware, show a colour pattern as start-up indicator.
 * @return void
 */
void Lightring::setup()
{
  pinMode(HW::PIN_LEDSTRIP, OUTPUT);
  strip = Adafruit_NeoPixel(HW::LEDSTRIP_LENGTH, HW::PIN_LEDSTRIP,
    HW::LEDSTRIP_TYPE);

  strip.begin();
  strip.setBrightness(0);
  strip.show();

  _timer_heartbeat.start(heartbeat_delay_ms);

  // Show a colour pattern as start-up indicator, with index brightness.
  if (Lightring::start_indicator_ms > 0) {
    strip.setBrightness(index_brightness);
    for (byte i = 0; i < HW::LEDSTRIP_LENGTH; ++i) {
      byte s = i * 0xC0/HW::LEDSTRIP_LENGTH;
      strip.setPixelColor(i, 0x80, s, 0xC0 - s);
    }
    strip.show();
    delay(start_indicator_ms);
    strip.setBrightness(0);
    strip.show();
  }
}

/*! @brief
 *
 * Depending on the setting mode, show the heart beat, ramp up/down the light's
 * brightness, or show whatever is appropriate for the setting mode.
 * @return void
 */
void Lightring::loop()
{
  // In setup mode light is currently set with each iteration, not only when
  // there was a change.
  if (_in_setup == SET_MENU) {
    _show_index(settings.val.menu_index);
  }
  else if (_in_setup == SET_BRIGHTNESS || _in_setup == SET_COLOUR_POS) {
    // Turn light on to current brightness and colour.
    _brightness = settings.val.brightness;
    _show_light();
  }
  else if (_in_setup == SET_ON_TIME) {
    _show_index(settings.val.on_time_index);
  }
  else if (_in_setup == SET_ON_RAMP_TIME) {
    _show_index(settings.val.on_ramp_index);
  }
  else if (_in_setup == SET_OFF_RAMP_TIME) {
    _show_index(settings.val.off_ramp_index);
  }
  else if (_in_setup == SET_HB_BRIGHTNESS) {
    _show_heartbeat();
  }
  else if (_in_setup == SET_COLOUR_LVL) {
    _show_wheel();
  }
  else if (_in_setup == SET_AMBIENT_BRIGHTNESS) {
    //_show_ambient();
  }
  else {
    if (_brightness != _target_brightness) {
      _ramp();
    }

    if (_brightness == 0) {
      _show_heartbeat();
    }
  }
}

/*! @brief
 *
 * Turn the light on or off.
 * Turn on ramping (up or down), and set target brightness (0 or full).
 * @param onoff bool True to turn the light on.
 * @return void
 */
void Lightring::setlight(bool onoff)
{
  if (onoff != _turn_on) {
    _ramping = true;
    _turn_on = onoff;
    _target_brightness = (_turn_on ? settings.val.brightness : 0);
  }
}

/*! @brief
 *
 * Ramp brightness up or down. The brightness step size is 1, the duration for
 * each step is calculated to go from start to finish within the set time
 * (on or off ramp duration).
 * @return void
 */
void Lightring::_ramp()
{
  // Check timer. Return if ramping and not yet expired.
  if (! _ramping || _timer_ramp.is_expired()) {
    _ramping = true;

    // Step brightness towards target.
    byte steplength;
    uint16_t ms;
    if (_brightness < _target_brightness) {
      _brightness++;
      ms = Lightring::get_ramp_time_choice_ms(settings.val.on_ramp_index);
    }
    else {
      _brightness--;
      ms = Lightring::get_ramp_time_choice_ms(settings.val.off_ramp_index);
    }
    steplength = ms / settings.val.brightness;

    // Update LED string.
    _show_light();

    // Set timer.
    _timer_ramp.startN(steplength);
  }
}

/*! @brief
 *
 * @return void
 */
void Lightring::_show_light()
{
  strip.setBrightness(_brightness);
  for (byte i = 0; i < HW::LEDSTRIP_LENGTH; ++i) {
    strip.setPixelColor(i, settings.val.colour);
  }
  strip.show();
}

/*! @brief
 *
 * Use the light as display for a 1-out-of-N selection.
 * The selection is zero-based and limited by the number of LEDs in the ring.
 * While in the setting menu (from which setting items are selected) use
 * colour idxcolour_menu, otherwise idxcolour_value.
 * With higher numbers blink the LED slower.
 * @param index byte The currently active selection. Zero-based.
 * @return void
 */
void Lightring::_show_index(byte index)
{
  if (index != _last_index  ||  _timer_idx_flash.is_expired()) {
    // index_offset is the LED used for index = 0.
    byte disp_index = index + index_offset;
    if (disp_index >= HW::LEDSTRIP_LENGTH)
      disp_index -= HW::LEDSTRIP_LENGTH;

    strip.setBrightness(index_brightness);
    for (byte i = 0; i < HW::LEDSTRIP_LENGTH; ++i) {
      uint32_t colour;
      colour = _in_setup == SET_MENU ? idxcolour_menu : idxcolour_value;
      colour = (i == disp_index && !_idx_pause) ? colour : 0;
      strip.setPixelColor(i, colour);
    }
    strip.show();

    // Blink slower with higher indices.
    // Should be calculated on max index, so far have only on_time_choice_size.
    _timer_idx_flash.start(_idx_pause ? 40 : (150 + index * 80));
    _idx_pause = !_idx_pause;
    _last_index = index;
  }
}

/*! @brief
 *
 * Show a colourwheel with the currently set brightness and level.
 * See ColourWheelRGBWCMY() about the level parameter.
 * @return void
 */
void Lightring::_show_wheel()
{
  strip.setBrightness(settings.val.brightness);
  for (byte i = 0; i < HW::LEDSTRIP_LENGTH; ++i) {
    byte idx = i * (255 / HW::LEDSTRIP_LENGTH);
    strip.setPixelColor(i, ColourWheelRGBWCMY(idx, settings.val.wheel_level));
  }
  strip.show();
}

/*! @brief
 *
 * Turn the next LED on when the interval has passed.
 * When in setting mode for heartbeat brightness, update continuously.
 * @return void
 */
void Lightring::_show_heartbeat()
{
  // Update constantly while in setting mode.
  if (_in_setup == SET_HB_BRIGHTNESS) {
    strip.setBrightness(settings.val.hb_brightness);
    strip.setPixelColor(_heartbeat_nr, heartbeat_colour);
    strip.show();
  }

  // Move to the next pixel each interval.
  if (_timer_heartbeat.is_expired()) {
    _timer_heartbeat.start(heartbeat_interval_ms);

    strip.setBrightness(settings.val.hb_brightness);
    strip.setPixelColor(_heartbeat_nr, 0); // Turn current pixel off.

    ++_heartbeat_nr;
    if (_heartbeat_nr >= HW::LEDSTRIP_LENGTH)
      _heartbeat_nr = 0;

    strip.setPixelColor(_heartbeat_nr, heartbeat_colour); // Turn next pixel on.
    strip.show();
  }
}

/*! @brief
 *
 * As long as this only shows something static it can be called from
 * setup_begin() instead of loop().
 * @return void
 */
void Lightring::_show_ambient()
{
    strip.setBrightness(index_brightness);
    for (byte i = 0; i < HW::LEDSTRIP_LENGTH; ++i) {
      strip.setPixelColor(i, ambient_colour);
    }
    strip.show();
}

/*! @brief
 *
 * Use the light for setting mode from now on.
 * Start with the light off.when setting heatbeat brightness.
 * @param setting_type byte The seting mode to enter.
 * @return void
 */
void Lightring::setup_begin(byte setting_type)
{
  _in_setup = setting_type;

  if (_in_setup == SET_HB_BRIGHTNESS) {
    // Turn light off. heartbeat() assumes it's off when it starts.
    _brightness = 0;
    _show_light();
  }
  else if (_in_setup == SET_AMBIENT_BRIGHTNESS) {
    _show_ambient();
  }
}

/*! @brief
 *
 * Use the light for normal operation from now on.
 * Return it to the same state as when setting mode was entered.
 * @return void
 */
void Lightring::setup_end()
{
  if (_in_setup == SET_BRIGHTNESS || _in_setup == SET_COLOUR_POS) {
    // Set light to what it was before entering setup.
    _brightness = _target_brightness;
  }

  _in_setup = SET_NONE;

  // Restore light to non-setting-mode state.
  _show_light();
}

/*! @brief
 *
 * Calculate the colour for a colour wheel position (0-255) and level (0..255).
 * This function produces colours for both RGB and CMY wheels by using the
 * level parameter.
 * Level: 0 - RGB wheel, 128 - white, 255 - CMY wheel.
 * Morph from RGB through white to CMY.
 * From 255 to 0 flip back from CMY to RGB.
 * @param index uint8_t Colour wheel position (0..255).
 * @param level uint8_t Colour wheel level (0..255).
 * @return uint32_t Colour value suitable for Adafruit_NeoPixel.
 */
uint32_t Lightring::ColourWheelRGBWCMY(uint8_t index, uint8_t level)
{
  uint8_t r, g, b, s;
  if (index < 1*255/3) {
    // Ramp down R, ramp up G, B at level.
    // 0:   0..255
    // 64:  64..191
    // 128: 128
    // 192: 191..64
    // 255: 255..0
    s = 128 + (((3 * index) - 128) * (128 - level)) / 128;
    r = 255 - s; g = s; b = level;
  }
  else if (index < 2*255/3) {
    // R at level, ramp down G, ramp up B.
    index -= 1*255/3;
    s = 128 + (((3 * index) - 128) * (128 - level)) / 128;
    r = level; g = 255 - s; b = s;
  }
  else {
    // Ramp up R, G at level, ramp down B.
    index -= 2*255/3;
    s = 128 + (((3 * index) - 128) * (128 - level)) / 128;
    r = s; g = level; b = 255 - s;
  }
  return Adafruit_NeoPixel::Color(r, g, b);
}

/*! @brief
 *
 * Load the settings from EEPROM. If the checksum is incorrect, load default
 * values.
 * @return void
 */
void NVSettings::load()
{
#ifdef DEBUG_APP
  _EEPROMdump('R');
#endif
  EEPROM.get(EEPROMaddr, val);
  if (! is_good_checksum()) {
#ifdef DEBUG_APP
    Serial.println("EE defaults");
#endif
    // Load defaults.
    val = SettingsDefaults;

    // Update the in-memory checksum.
    // Not really needed, it's always updated before writing to EEPROM.
    checksum_update();
  }
}

/*! @brief
 *
 * Update the checksum and write the settings to EEPROM.
 * @return void
 */
void NVSettings::store()
{
  // Update checksum before writing to EEPROM.
  checksum_update();
  EEPROM.put(EEPROMaddr, val);
#ifdef DEBUG_APP
  _EEPROMdump('W');
#endif
}

/*! @brief
 *
 * @return bool True if the checksum for the settings data is correct.
 */
bool NVSettings::is_good_checksum()
{
  uint16_t s = 0;
  for (byte i = 0; i < valsize; ++i)
    s += ((byte *)&val)[i];
  _checksum = s;
#ifdef DEBUG_APP
  Serial << "EE sum: 0x"; Serial.println(_checksum, HEX);
#endif

  // Checksum is correct if all bytes, incl checksum, sum up
  // to _good_checksum_value.
  return _checksum == _good_checksum_value;
}

/*! @brief
 *
 * Update the checksum for the settings in memory.
 * After this, is_good_checksum() returns true.
 * @return void
 */
void NVSettings::checksum_update()
{
  // Calculate current sum.
  is_good_checksum();

  // Correct the checksum.
  // The checksum byte(s) were already summed up; subtract first.
  _checksum -= val.checksum;
  // Add what's missing to get sum of _good_checksum_value.
  _checksum = _good_checksum_value - _checksum;
#ifdef DEBUG_APP
  Serial << "EE sum new: 0x";  Serial.println(_checksum, HEX);
#endif

  val.checksum = _checksum;
}

#ifdef DEBUG_APP
/*! @brief
 * Dump the EEPROM memory used for settings storage. Debugging only.
 * @param id char Character to show for from where this is called.
 * @return void
 */
void NVSettings::_EEPROMdump(char id)
{
  Serial << "EE: " << id << endl;
  for (byte i = EEPROMaddr; i < EEPROMaddr+valsize; ++i) {
    Serial << i << "\tx";
    Serial.println(EEPROM.read(i), HEX);
  }
}
#endif

/*! @brief
 *
 * @return void
 */
void Settingup::setup()
{
}

/*! @brief
 *
 * Handle all the user input (rotary encoder, push buttons).
 * Change state between normal and setting mode.
 * @return void
 */
void Settingup::loop()
{
  if (_setting_up == SET_MENU) {
    settings.val.menu_index = _set_range(settings.val.menu_index,
          SET_NUM_ITEMS - 1);
    _check_enter_setup();
  }
  else if (_setting_up == SET_BRIGHTNESS) {
    settings.val.brightness = _set_range(settings.val.brightness,
          Lightring::on_brightness_min, Lightring::on_brightness_max);
    _set_brightness(); // Old push buttons.
  }
  else if (_setting_up == SET_ON_TIME) {
    settings.val.on_time_index = _set_range(settings.val.on_time_index,
          Lightring::on_time_choice_size - 1);
    _set_on_time(); // Old push buttons.
  }
  else if (_setting_up == SET_ON_RAMP_TIME) {
    byte val = settings.val.on_ramp_index;
    val = _set_range(val, Lightring::ramp_time_choice_size - 1);
    settings.val.on_ramp_index = val;
  }
  else if (_setting_up == SET_OFF_RAMP_TIME) {
    byte val = settings.val.off_ramp_index;
    val = _set_range(val, Lightring::ramp_time_choice_size - 1);
    settings.val.off_ramp_index = val;
  }
  else if (_setting_up == SET_HB_BRIGHTNESS) {
    byte val = settings.val.hb_brightness;
    val = _set_range(val, Lightring::heartbeat_brightness_min,
                      Lightring::heartbeat_brightness_max);
    settings.val.hb_brightness = val;
  }
  else if (_setting_up == SET_COLOUR_POS || _setting_up == SET_COLOUR_LVL) {
    _set_colour();
  }
  else if (_setting_up == SET_AMBIENT_BRIGHTNESS) {
    _set_ambient_night();
  }
  else {
    _check_enter_setup();
  }

  _check_exit_setup();

#ifdef DEBUG_APP
  // Call event functions to show events. The events are lost.
  // Only call those here that aren't called elsewhere!
#ifdef SUPPORT_UPDOWN_BUTTONS
  if (_setting_up == SET_NONE) {
    button_down.is_single();
    button_up.is_single();
  }
  button_down.is_released();
  button_up.is_released();
#endif
//  encoder.sw.is_single();
  encoder.sw.is_long_single();
  encoder.sw.is_released();
  if (_setting_up == SET_NONE) {
    static int cnt, lastcnt = -1;
    cnt += encoder.change();
    if (cnt != lastcnt) {
      Serial << "encoder: \t" << cnt << endl;
      lastcnt = cnt;
    }
  }
#endif
}

/*! @brief
 *
 * @return bool True if in setting mode.
 */
bool Settingup::in_setup_mode()
{
  return _setting_up != SET_NONE;
}

/*! @brief
 *
 * Read inputs and enter setting mode, or the selected setting item.
 * @return void
 */
void Settingup::_check_enter_setup()
{
  byte settype = SET_NONE;

  // Enter setup mode or select setup item with rotary encoder push switch.
  if (encoder.sw.is_single()) {
    if (_setting_up == SET_NONE) {
      settype = SET_MENU;

      // Always enter the setting up menu at the lowest index.
      // Otherwise, entry is at the last used index.
      //settings.val.menu_index = 0;
    }
    else if (_setting_up == SET_MENU) {
      settype = settings.val.menu_index + SET_MENU_FIRST;
#ifdef SUPPORT_UPDOWN_BUTTONS
      _entered_by_updown = false;
#endif
    }
  }

#ifdef SUPPORT_UPDOWN_BUTTONS
  // Enter setup mode with long button press.
  if (button_down.is_long_single()) {
    settype = SET_BRIGHTNESS;
    _entered_by_updown = true;
  }
  else if (button_up.is_long_single()) {
    settype = SET_ON_TIME;
    _entered_by_updown = true;
  }
#endif

  if (settype != SET_NONE) {
#ifdef DEBUG_APP
    Serial << "setup start " << settype << endl;
#endif
    _setting_up = settype;
    statusdisp.setting_mode(_setting_up);
    light.setup_begin(_setting_up);
    _setup_timeout.start(setup_timeout_ms);
  }
}

/*! @brief
 *
 * Exit setting mode after inactivity.
 * @return void
 */
void Settingup::_check_exit_setup()
{
  if (_setting_up != SET_NONE) {
    // Re-start timeout as long as there is activity.
    // Holding a button down is activity and shouldn't exit setting mode.
    if (encoder.sw.is_pressed_raw()
#ifdef SUPPORT_UPDOWN_BUTTONS
        || button_down.is_pressed_raw() || button_up.is_pressed_raw()
#endif
       )
      _setup_timeout.restart();

    if (_setup_timeout.is_expired()) {
#ifdef DEBUG_APP
      Serial << "setup exit " << _setting_up << "  ";
      _exit_debug();
      Serial << endl;
#endif
      if (_setting_up == SET_MENU
#ifdef SUPPORT_UPDOWN_BUTTONS
      	  || _entered_by_updown
#endif
          ) {
        // In settings menu - exit setup mode.
        _setting_up = SET_NONE;
        light.setup_end();
        settings.store();
        statusdisp.setting_mode(SET_NONE);
      }
      else {
        // Go back to settings menu.
        _setting_up = SET_MENU;
        statusdisp.setting_mode(SET_MENU);
//        light.setup_end(); // Allow light object to tidy up.
        light.setup_begin(_setting_up);
        _setup_timeout.restart();
      }
    }
  }
}

/*! @brief
 *
 * @return void
 */
void Settingup::_exit_debug()
{
#ifdef DEBUG
  Serial << "new: ";
  switch (_setting_up) {
    case SET_MENU:
      Serial << "menu " << settings.val.menu_index;
      break;
    case SET_BRIGHTNESS:
      Serial << "bright " << settings.val.brightness;
      break;
    case SET_ON_TIME:
      Serial << "on-time "
        << Lightring::get_on_time_choice(settings.val.on_time_index) << 's';
      break;
    case SET_ON_RAMP_TIME:
      Serial << "ramp on "
        << Lightring::get_ramp_time_choice_ms(settings.val.on_ramp_index)
        << "ms";
      break;
    case SET_OFF_RAMP_TIME:
      Serial << "ramp off "
        << Lightring::get_ramp_time_choice_ms(settings.val.off_ramp_index)
        << "ms";
      break;
    case SET_HB_BRIGHTNESS:
      Serial << "HB bright " << settings.val.hb_brightness;
      break;
    case SET_COLOUR_POS:
    case SET_COLOUR_LVL: {
      uint8_t pos = settings.val.wheel_position;
      uint8_t lvl = settings.val.wheel_level;
      uint32_t colour = Lightring::ColourWheelRGBWCMY(pos, lvl);
      Serial << "pos " << pos << " lvl " << lvl << " colour " << _HEX(colour);
      break;
    }
    case SET_AMBIENT_BRIGHTNESS:
      Serial << "ambient " << settings.val.night_threshold_ADC;
      break;
  }
#endif
}

/*! @brief
 *
 * User selection of a range between 0 and max.
 * @param current byte Current selection. Zero-based.
 * @param max byte Highest possible selection.
 * @return byte New selection. Zero-based.
 */
byte Settingup::_set_range(byte current, byte max)
{
  return _set_range(current, 0, max);
}

/*! @brief
 *
 * @param current byte Current selection. Zero-based.
 * @param min byte Lowest possible selection.
 * @param max byte Highest possible selection.
 * @return byte New selection. Zero-based.
 */
byte Settingup::_set_range(byte current, byte min, byte max)
{
  int16_t change = encoder.change();
  if (change != 0) {
    change += current;

    if (change < min)
      change = min;
    if (change > max)
      change = max;
    current = change;
#ifdef DEBUG_APP
    Serial << "range new val: " << current << endl;
#endif

    // Restart activity timeout.
    _setup_timeout.restart();
  }

  return current;
}

/*! @brief
 *
 * Select a colour by colour wheel position and value.
 * The selected colour is shown at the currently selected brightness when
 * the light is on.
 * Position and level are two different setting items, both handled here.
 * @return void
 */
void Settingup::_set_colour()
{
  // Pushing knob changes between setting colour wheel index and colour
  // wheel level.
  if (encoder.sw.is_single()) {
    _setting_up =
      (_setting_up == SET_COLOUR_POS) ? SET_COLOUR_LVL : SET_COLOUR_POS;
    light.setup_begin(_setting_up);
  }

  uint8_t cwpos = settings.val.wheel_position;
  uint8_t cwlevel = settings.val.wheel_level;

  int8_t change = encoder.change();
  if (change != 0) {
    if (_setting_up == SET_COLOUR_LVL) {
      // Colour wheel level. Stop at 255 as well: 248 255 0 8, 8 0 255 248.
      if (cwlevel == 256-8 && change > 0  ||  cwlevel == 0 && change < 0)
        cwlevel = 255;
      else if (cwlevel == 255 && change < 0)
        cwlevel = 256-8;
      else {
        cwlevel += 8 * change;
        cwlevel &= 0xF8;
      }
    }
    else {
      // Colour wheel position. Keep multiples of 5: 250 255 5, 5 0 250.
      if (cwpos == 255)
        cwpos++; // make 1 step of 85 1/5 larger (going up)
      cwpos += change * 5;
      if (cwpos == 251)
        cwpos--; // make 1 step of 85 1/5 larger (going down)
    }
    settings.val.wheel_position = cwpos;
    settings.val.wheel_level = cwlevel;

    settings.val.colour = Lightring::ColourWheelRGBWCMY(
      settings.val.wheel_position, settings.val.wheel_level);
#ifdef DEBUG_APP
    uint8_t r, g, b;
    r = settings.val.colour >> 16;
    g = settings.val.colour >> 8;
    b = settings.val.colour;
    Serial << "wheel: " << cwpos << '\t' << r << ',' << g << ',' << b
      << " \tl3 " << cwlevel << endl;
#endif

    _setup_timeout.restart();
  }
}

void Settingup::_set_ambient_night()
{
  settings.val.night_threshold_ADC = daylight.light_value();
#ifdef DEBUG_APP
  static SoftTimer showval;
  if (showval.interval(1000/7))
    Serial << "Light ADC " << settings.val.night_threshold_ADC
      << "  adc " << daylight.ADCval() << endl;
#endif
}

/*! @brief
 *
 * Change brightness with the old two push buttons. Their presence is optional.
 * @return void
 */
void Settingup::_set_brightness()
{
#ifdef SUPPORT_UPDOWN_BUTTONS
  if (button_down.is_single()
      && settings.val.brightness >= Lightring::on_brightness_min)
    settings.val.brightness -= Lightring::on_brightness_change;
  if (button_up.is_single()
      && settings.val.brightness <=
          Lightring::on_brightness_max - Lightring::on_brightness_change)
    settings.val.brightness += Lightring::on_brightness_change;
#endif
}

/*! @brief
 *
 * Change on-time with the old two push buttons. Their presence is optional.
 * @return void
 */
void Settingup::_set_on_time()
{
#ifdef SUPPORT_UPDOWN_BUTTONS
  if (button_down.is_single()) {
    --settings.val.on_time_index;
    if (settings.val.on_time_index >= Lightring::on_time_choice_size)
      settings.val.on_time_index = 0;
    // Note: -1 shows as max in unsigned values.
    //Serial.print("ON idx "); Serial.println(settings.val.on_time_index);
  }
  if (button_up.is_single()) {
    ++settings.val.on_time_index;
    if (settings.val.on_time_index >= Lightring::on_time_choice_size)
      settings.val.on_time_index = Lightring::on_time_choice_size - 1;
    //Serial.print("ON idx "); Serial.println(settings.val.on_time_index);
  }
#endif
}


/*
 * Arduino program
 */

/*! @brief Blink the built-in LED 3x on start-up, to show when the CPU is reset.
 *
 * @return void
 */
void show_start_blinks()
{
  bool LEDon = false;
  for (byte i = 7; i > 0; --i) {
    digitalWrite(LED_BUILTIN, LEDon);
    // Start with off, or 1st blink might be invisible. Finish with off, ditto.
    LEDon = !LEDon;
    delay(100);
  }
}

/*! @brief Arduino program - setup.
 *
 * This function is called once after reset.
 * @return void
 */
void setup()
{
  // Set up light first, in case it's on from power-up, to turn it off fast.
  light.setup();

#ifdef SHOW_START_BLINKS
  show_start_blinks();
#endif

#if defined(SHOW_APP_VERSION) || defined(DEBUG_APP)
  Serial.begin(HW::RS232_speed, HW::RS232_mode);
  Serial.print("\r\nStartup...\r\n");
#endif

  // Load settings from EEPROM.
  settings.load();

  // Set up all remaining parts.
  statusdisp.setup();
#if defined(SHOW_APP_VERSION) || defined(DEBUG_APP)
  Serial.print(FS(APPLICATIONSUMMARY_TEXT));
#endif
  pir.setup();
  daylight.setup();
  settingup.setup();
}

/*! @brief Arduino program - loop.
 *
 * This function is called in an endless loop.
 * Poll all the system parts and the switches.
 * @return void
 */
void loop()
{
  light.loop();
  statusdisp.loop();
  pir.loop();
  daylight.loop();
  settingup.loop();

  if (encoder.sw.loop_is_do_service()) {
    encoder.sw.service(); // For the encoder button.
#ifdef SUPPORT_UPDOWN_BUTTONS
    button_down.service();
    button_up.service();
#endif
  }

  // Shouldn't be called more often than every 1ms, but the program probably
  // takes that long anyway. Umph, 13200 - 17900 iterations/s... and 1ms is not
  // often enough and can loose count when spinning fast.
  if (encoder.loop_is_do_service()) {
    encoder.service(); // For the encoder A/B switches, not its button.
  }

#ifdef DEBUG_APP
  // Find out how fast we get around.
  if (millis() > 3000) { // Start after 3s
    static uint32_t ms;
    static uint16_t cnt = 0;
    if (cnt == 0) {
      ms = millis() + 1000;
      cnt++;
    }
    else if (millis() < ms)
      cnt++;
    else if (cnt < -1) {
      Serial << "ticks in 1s: " << cnt << endl;
      cnt = -1;
    }
  }
#endif

  // Show state of PIR pin on LED.
  // (If it shares LED with setting mode, only when not in setting mode.)
  statusdisp.PIRactive(pir.raw());

  // Turn light on when PIR triggered, off after timeout.
  if (! settingup.in_setup_mode()) {
    // Turn light on if PIR is activated and it is dark, off after timeout.
    // Re-trigger is permitted.
    static SoftTimer timer_light;
    static bool pir_triggered = false;
    if (pir.is_active() && daylight.is_night()) {
      unsigned long ms;
      ms = Lightring::get_on_time_choice(settings.val.on_time_index) * 1000;
      timer_light.startW(ms);
      light.setlight(true);
      pir_triggered = true;
    }
    if (pir_triggered) {
      if (timer_light.is_expired()) {
        light.setlight(false);
        pir_triggered = false;
      }
    }
  }
}

/******************************************************************************/

/***************************************************************************//**
@mainpage Night Light With PIR Manual

@section Features

Automatic night light with motion detector (PIR sensor).
The light is turned on slowly to be easier on the eyes, and is not very bright
to avoid waking up fully when getting up at night.
The light is turned off slowly to avoid being in the dark suddenly.
Light colour can be changed to what's best for night vision.
The light is not turned on at daytime (above a certain amount of ambient light).
When the light is off, a single dim LED walks around the ring to indicate that
the program is still running (heartbeat).

Adjustments:
- Brightness.
- Length of time for which the light is on.
- Time over which the light reaches full brightness.
- Time over which the light goes off.
- Heartbeat brightness. Set to 0 to turn off.
- Light colour.
- Ambient light level above which the light is not turned on.

@sa night-light-PIR.ino (details)

@section Operation

When the PIR is triggered, turn on light slowly to be easy on the eyes.
After timeout, turn off light slowly, to avoid being in the dark suddenly.
The light can be re-triggered, and the configured on-time is counted from the
last time the PIR was triggered.

Enter setup by pushing the knob. Push the knob again to enter setup item.
Blue LEDs show setup menu, green ones are a value selection (in setup item).
- Setup index 1: Light brightness.
- Setup index 2: Light on-time.
- Setup index 3: How fast the light turns on.
- Setup index 4: How fast the light turns off.
- Setup index 5: Heartbeat brightness.
- Setup index 6: Light colour (colour wheel position).
- Setup index 7: Light colour (colour wheel level).
- Setup index 8: Ambient light level threshold day/night.

The index light blinks slower for higher indices / longer times.
If there is no input for 4s it  returns to the setup menu, no input for another
4s exits setup and stores new values in non-volatile memory.
Pushing the knob while setting light colour toggles between position and level.

The program version etc. are output to the serial port (115200 8N1) if that
was enabled at compile time.

Version 1 up/down buttons:
The first prototype also had 2 PCB-mounted push buttons, "up" and "down".
They are still supported if present, and can be used to change brightness and
on-time.
Enter setup by holding down-button for 1s; light is turned on, adjust with
up/down.
Enter setup by holding up-button for 1s; select from 8 different on-times
(the time the light is on after being triggered), ranging from 15s to 16min.
The index light blinks slower for longer on-times.
Exit setup after 4s without button press.
The chosen brightness and on-time are stored in non-volatile memory and are
unchanged after power cycling.

Developed and tested with Arduino IDE 1.6.10.
Does not compile on 1.5.X or earlier, but would probably only require
minor changes (EEPROM doesn't have put()).


@section implementation Implementation outline

All system parts are implemented using C++ objects, and use state machines and
polling to be asynchronous and to exit as fast as possible. This keeps the
system responsive without having to use interrupts. It's a kind of
multi-tasking. Akin to the Arduino, the object methods are called setup() and
loop().

The rotary encoder turned out to produce results in reverse for turning. Whether
that is due to encoder construction or the PushButton library is unknown. The
easy way to address it is to swap the pin configurations for the encoder's A and
B signals.

The prototype was built with an Arduino Pro Mini 16MHz 5V ATmega168.
Many other Arduinos will work too, as long as they have at least 16kbyte of
flash memory. It does not quite fit into an 8kbyte ATtiny (with 6kbyte free).

Memory use (v1.6):
- No button support and no use of serial port:
Flash 6,372 bytes, RAM 253 bytes.
- Up/down button support:
Flash 6,648 bytes, RAM 278 bytes.
- Printing app name and version to serial port:
Flash 7,828 bytes, RAM 467 bytes.
- With debugging output:
Flash 9,106 bytes, RAM 650 bytes.


@section Author

Volker Kuhlmann
@n
http://volker.top.geek.nz/contact.html


@section Download

http://volker.top.geek.nz/arduino/


@section Copyright

Copyright (C) 2016 by Volker Kuhlmann  @n
	      http://volker.top.geek.nz/contact.html  @n
All rights reserved.

Night-light-PIR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Night-light-PIR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Night-light-PIR.  If not, see <http://www.gnu.org/licenses/>.


@section TODO
- Nothing at the moment.


@section History
- v1.6, 19 Sep 2016  Volker Kuhlmann
  - Released with doxygen comments and documentation.
  - GPLv3

*******************************************************************************/
